/*
 * Servocontroller 1.6
 * © Fenna Pel, fenna@pereltje.nl
 * 
 * gebaseerd op dit artikel:
 * https://learn.adafruit.com/multi-tasking-the-arduino-part-2/overview by Bill Earl
 * 
 * De code is aangepast en deels herschreven om tot 7 servo's te kunnen aansturen met instelbare snelheid en uitslag.
 * Om de aanpassingen heen is nieuwe code geschreven om de speciefieke eigenschappen van de gebruikte servo's netjes op te vangen en om functionaliteit toe te voegen.
 * 
 * voor versie 1.6
 *   - 
 * 
 * voor versie 1.5 (2020-11-17) 
 *   - code verder opgeschoond
 *   - 'int' types voor 'short' vervangen waar mogelijk, voorsorteren op de opslag van de waarden in de EEPROM en zuiniger op het RAM gebruik
 *   - servo aansturing via array's ipv gewone variabelen
 *   - Een aantal hardgecodeerde constantes voor systeem variabelen vervangen
 *   - Eerste aanzet voor range programeren (placeholder subroutines aangemaakt voor lezen/schrijven EEPROM)
 *   - code terug geschreven voor seriele monitor
 *   - Een servo monitor subroutine toegevoegd voor seriele monitor
 *   - Bij elke timer interrupt (1/ms) wordt slechts 1 servo bijgesteld ipv alle 7 (geeft soepelere slag op alle servo's) Een teller houdt bij welke servo aan de beurt is
 *   - 
 *   
 * voor versie 1.4
 * 
 * voor versie 1.3 (2020-10-31)
 *   - code opgeschoond en hard gecodeerde waarden voor pinnen omgeschreven naar vooraf gedefinieerde variabelen/constanten
 *   - code klaargemaakt voor 7 servo's ipv de 2 hard gecodeerde servo's
 *   - gracefull shutdown geeft nu een knipperende power led en is in een aparte routine ondergebracht
 *   - verwijzingen en code restjes voor seriele monitor verwijderd. voor een monitor kan evt een I2C display worden ingeprogrammeerd
 *   
 */



#include <Servo.h> 
#include <Wire.h>
#include <Adafruit_ADS1015.h>
#include <EEPROM.h>

// systeem variabelen en constanten
const unsigned short _SERVORANGE_MIN = 1000;
const unsigned short _SERVORANGE_MAX = 2000;
const unsigned short _SERVORANGE = _SERVORANGE_MAX - _SERVORANGE_MIN;
const unsigned short _SERVOCENTRE = (_SERVORANGE_MIN + _SERVORANGE_MAX)/2;
short _BOARDPOWER; // wordt ook gebruikt in de sweeper class
const short _SHUTDOWN_DELAY = 10000;
const short _START_DELAY = 500;
const short _INCREMENT = 3;
const short _REVERSE_DELAY = 3000;
unsigned short _INTERRUPT_COUNTER = 0;
const short _ADA_ADC_5V = 1695; //deze waarde is mogelijk individueel hardware afhankelijk van de gebruikte I2C ADC. Met de servo print routine is de waarde te achterhalen
const short _ARD_ADC_5V = 1023; //vast gegeven voor hoe de arduino eigen ADC's werken. Voor code consistentie via deze systeem constante
const unsigned short _SWEEP_MIN = 1;
const unsigned short _SWEEP_MAX = 50;
const short _ADS1015_I2C_1 = 0x48; //I2C adres van de gebruikte adafruit I2C ADC

// EEPROM handling:
byte _EEPROM_BLOCK_COPY[32] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
//unsigned short _EEPROM_RANGE[7] = {0,0,0,0,0,0,0};
unsigned short _EEPROM_POSITION[7] = {_SERVOCENTRE,_SERVOCENTRE,_SERVOCENTRE,_SERVOCENTRE,_SERVOCENTRE,_SERVOCENTRE,_SERVOCENTRE};
unsigned int _EEPROM_WRITE_COUNT; // Keep track of the number of write cycles to the EEPROM
byte _EEPROM_BLOCK_NUMBER; // Keep track of which EEPROM block is used
int _EEPROM_BLOCK_START_ADDRESS;
const int _EEPROM_MAX_WRITES = 90000;
int updatedposition;



// pinverdeling
short potmeterpin[7] = {A0,A1,A2,A3,0,1,2};
short servopin[7] = {6,7,8,9,10,11,12};
short potmeterpin8 = 3;
short ledpin = 13; //powerled
short ledadr0 = 3; // deze 3 pins gaan naar 3-8 decoder voor indicatie bij welke servo de range wordt ingesteld
short ledadr1 = 4; //
short ledadr2 = 5; //
short poweroffpin = A7;
short programpin = A6;
short servoonpin = 2;

//  variabelen met hun startstanden:
short potmeter[7] = {0,0,0,0,0,0,0};;
short potmeter8 = 0;
short timeoff_Pon = 1;
short timeoff_Poff = 250;
short timeon = 250;
short ledadr = 7;
short programkeyvalue = 0;
short programkey = 0;
short programkey2 = 0;
unsigned short servospeed[7] = {_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN};
unsigned short servorange[8] = {_SERVORANGE_MAX,_SERVORANGE_MAX,_SERVORANGE_MAX,_SERVORANGE_MAX,_SERVORANGE_MAX,_SERVORANGE_MAX,_SERVORANGE_MAX,0};



void gracefullshutdown()
{
  
  digitalWrite(ledadr0, HIGH);
  digitalWrite(ledadr1, HIGH);
  digitalWrite(ledadr2, HIGH);
  
  //increment write counter of the current block, max out at _EEPROM_MAX_WRITES and store at beginning of the current block
  _EEPROM_WRITE_COUNT = EEPROM.read(_EEPROM_BLOCK_START_ADDRESS+1)*65536 + EEPROM.read(_EEPROM_BLOCK_START_ADDRESS+2)*256 + EEPROM.read(_EEPROM_BLOCK_START_ADDRESS+3);
  _EEPROM_WRITE_COUNT++;
  if ( _EEPROM_WRITE_COUNT > _EEPROM_MAX_WRITES) // prepare for a new block in EEPROM
  {
    _EEPROM_BLOCK_NUMBER++;
    _EEPROM_WRITE_COUNT = 0;
    _EEPROM_BLOCK_START_ADDRESS = _EEPROM_BLOCK_NUMBER * 32;
  }
  EEPROM.update(0, _EEPROM_BLOCK_NUMBER);
  EEPROM.update(_EEPROM_BLOCK_START_ADDRESS+1, _EEPROM_WRITE_COUNT/65536);
  EEPROM.update(_EEPROM_BLOCK_START_ADDRESS+2, (_EEPROM_WRITE_COUNT%65536)/256);
  EEPROM.update(_EEPROM_BLOCK_START_ADDRESS+3, _EEPROM_WRITE_COUNT%256);

  // store servo values:
  for (byte i = 0; i <= 6; i++) 
  {
    int base_address = ((1+i)*4+_EEPROM_BLOCK_START_ADDRESS);
    EEPROM.update(base_address, _EEPROM_POSITION[i]/256);
    EEPROM.update(base_address+1, _EEPROM_POSITION[i]%256);
    EEPROM.update(base_address+2, servorange[i]/256);
    EEPROM.update(base_address+3, servorange[i]%256);
  }
  Serial.println();
  Serial.println(_EEPROM_WRITE_COUNT);
  delay(_SHUTDOWN_DELAY);
  digitalWrite (ledpin, LOW);
  digitalWrite(servoonpin, LOW);
}

void printservo(short servonr)
{
  Serial.print ("servo: ");
  Serial.print (servonr);
  Serial.print (" , potmeterstand: ");
  Serial.print (potmeter[servonr-1]);
  Serial.print (" snelheid: ");
  Serial.print (servospeed[servonr-1]);
  Serial.print (" range: ");
  Serial.print (servorange[servonr-1]);
  Serial.print (" eeprom position: ");
  Serial.println (_EEPROM_POSITION[servonr-1]);
}

// setup routine:
void setup() 
{
  Serial.begin(9600);
  _EEPROM_BLOCK_NUMBER = EEPROM.read(0); // determine which block is used and calculate block offset (the current blocknumber is stored at the first EEPROM address)
  _EEPROM_BLOCK_START_ADDRESS = _EEPROM_BLOCK_NUMBER * 32; 
  Serial.println(_EEPROM_BLOCK_NUMBER);
  // get most recent values from EEPROM

    // store servo values:
  for (byte i = 0; i <= 6; i++) 
  {
    int base_address = ((1+i)*4+_EEPROM_BLOCK_START_ADDRESS);
    EEPROM.update(base_address, _EEPROM_POSITION[i]/256);
    EEPROM.update(base_address+1, _EEPROM_POSITION[i]%256);
    EEPROM.update(base_address+2, servorange[i]/256);
    EEPROM.update(base_address+3, servorange[i]%256);
  }

  for (byte i = 0; i <= 6; i++) 
  {
     int base_address = ((1+i)*4+_EEPROM_BLOCK_START_ADDRESS);
     _EEPROM_POSITION[i] = EEPROM.read(base_address) * 256 + EEPROM.read(base_address+1);
     //if (_EEPROM_POSITION[i] == 0)  _EEPROM_POSITION[i]=_SERVOCENTRE;
    servorange[i] =  EEPROM.read(base_address+2) * 256 + EEPROM.read(base_address+3);
   
    Serial.print(base_address);
    Serial.print (": ");
    Serial.print (EEPROM.read(base_address));
    Serial.print (" ");
    Serial.print (EEPROM.read(base_address+1));
    Serial.print (" ");
    Serial.print (EEPROM.read(base_address+2));
    Serial.print (" ");
    Serial.print (EEPROM.read(base_address+3));
    Serial.print (" -- ");
    Serial.print (_EEPROM_POSITION[i]);
    Serial.print (" -- ");
    Serial.print (servorange[i]);
    Serial.println();
    
  
  }
  Serial.println();
  delay(_START_DELAY);


  
  // Timer0 is already used for millis() - we'll just interrupt somewhere
  // in the middle and call the "Compare A" function below
  OCR0A = 0xAF;
  TIMSK0 |= _BV(OCIE0A);
  
//gracefull start

pinMode(servoonpin, OUTPUT);
digitalWrite(servoonpin, HIGH);
pinMode(poweroffpin, INPUT);
pinMode(ledadr0, OUTPUT);
digitalWrite(ledadr0, HIGH);
pinMode(ledadr1, OUTPUT);
digitalWrite(ledadr1, HIGH);
pinMode(ledadr2, OUTPUT);
digitalWrite(ledadr2, HIGH);
} 



// hoofd programma:
void loop()
{
  
}
